/*
 * $Id: gt_node_list.c,v 1.1 2003/08/24 08:46:47 hipnod Exp $
 *
 * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "gt_gnutella.h"

#include "gt_node.h"
#include "gt_node_list.h"

/*
 * TODO: rename gt_conn -> gt_node_list
 */

/*****************************************************************************/

/* list of all nodes -- NOTE: duplicated info in gt_node.c */
static List      *node_list;

/* last place in node_list for gt_conn_foreach */
static List      *iterator;

/*****************************************************************************/

void gt_conn_add (GtNode *node)
{
	if (!node)
	{
		GIFT_ERROR (("adding null node to node list"));
		return;
	}

	node_list = list_append (node_list, node);

	/* reset the iterator to this node */
	if (!iterator)
		iterator = node_list;
}

void gt_conn_remove (GtNode *node)
{
	if (!list_find (node_list, node))
		return;

	if (list_nth_data (iterator, 0) == node)
		iterator = iterator->next;

	node_list = list_remove (node_list, node);

	if (!iterator)
		iterator = node_list;
}

static void trace_list (List *nodes)
{
	GtNode *node;

	if (!nodes)
		return;

	node = list_nth_data (nodes, 0);

	assert (node != NULL);
	assert (GT_CONN(node) != NULL);

	GT->DBGFN (GT, "%s:%hu", net_ip_str (node->ip), node->gt_port);

	if (list_next (nodes))
		trace_list (list_next (nodes));
}

/*****************************************************************************/

GtNode *gt_conn_foreach (GtConnForeachFunc func, void *udata,
                         GtNodeClass klass, GtNodeState state,
                         int iter)
{
	GtNode      *node;
	TCPC        *c;
	GtNode      *ret       = NULL;
	List        *ptr;
	List        *start;
	List        *next;
	unsigned int i, count;
	int          looped    = FALSE;
	int          iterating = FALSE;

	assert (func != NULL);

#if 0
	GT->DBGFN (GT, "length of conn list: %u", list_length (connections));
#endif

	if (iter)
		iterating = TRUE;

	if (!iterator)
		iterator = node_list;

	start = ptr = (iterating) ? iterator : node_list;

	/* having count be the static list length should keep
	 * concurrent conn_adds from making us never stop */
	count = list_length (node_list);

	/* hack for backward-compatible interface */
	if (state == (GtNodeState) -1)
		state = GT_NODE_ANY;

	for (i = 0; i < count; i++)
	{
		if (iter && !ptr && !looped)
		{
			/* data only gets appended to connection list:
			 * safe to use head of connection list (connections) */
			ptr = node_list;
			looped = TRUE;
		}

		if (!ptr)
			break;

		/* we wrapped to the beginning, but have just reached the original
		 * start so we should bail out */
		if (looped && ptr == start)
			break;

		node = ptr->data;
		c = GT_CONN(node);

		assert (node != NULL);

		if (klass && !(node->klass & klass))
		{
			ptr = list_next (ptr);
			continue;
		}

		if (state != GT_NODE_ANY && node->state != state)
		{
			ptr = list_next (ptr);
			continue;
		}

		/* grab the next item. this allows the callback to free this item */
		next = list_next (ptr);

		ret = (*func) (c, node, udata);

		ptr = next;

		if (ret)
			break;

		if (iterating && --iter == 0)
			break;
	}

	/* save the position for next time */
	if (iterating)
		iterator = ptr;

	return ret;
}

/*****************************************************************************/

static GtNode *conn_counter (TCPC *c, GtNode *node, int *length)
{
	(*length)++;
	return NULL;
}

int gt_conn_length (GtNodeClass klass, GtNodeState state)
{
	int ret = 0;

	gt_conn_foreach ((GtConnForeachFunc) conn_counter,
	                 &ret, klass, state, 0);

	return ret;
}

static GtNode *select_rand (TCPC *c, GtNode *node, void **cmp)
{
	int     *nr    = cmp[0];
	GtNode **ret   = cmp[1];
	float    range = *nr;
	float    prob;

	/* make sure we pick at least one */
	if (!*ret)
		*ret = node;

	/* set the probability of selecting this node */
	prob = range * rand() / (RAND_MAX + 1.0);

	if (prob < 1)
		*ret = node;

	(*nr)++;

	/* we dont use the return value here, because we need to try
	 * all the nodes, and returning non-null here short-circuits */
	return NULL;
}

/*
 * Pick a node at random that is also of the specified 
 * class and state.
 */
GtNode *gt_conn_random (GtNodeClass klass, GtNodeState state)
{
	void   *cmp[2];
	int     nr;   
	GtNode *ret = NULL;

	/* initial probability */
	nr = 1;

	cmp[0] = &nr;
	cmp[1] = &ret;

	gt_conn_foreach ((GtConnForeachFunc) select_rand,
	                 &cmp, klass, state, 0);

	return ret;
}

/*****************************************************************************/

int gt_conn_sort_vit (GtNode *a, GtNode *b)
{
	if (a->vitality > b->vitality)
		return -1;
	else if (a->vitality < b->vitality)
		return 1;
	else
		return 0;
}

/* NOTE: this isnt safe to call at all times */
void gt_conn_sort (CompareFunc func)
{
	node_list = list_sort (node_list, func);
}
