/*
 * $Id: rx_packet.c,v 1.3 2003/09/17 19:13:59 hipnod Exp $
 *
 * Copyright (C) 2003 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "gt_gnutella.h"

#include "rx_stack.h"
#include "rx_layer.h"

#include "gt_packet.h"
#include "rx_packet.h"

/*****************************************************************************/

struct rx_packet
{
	struct io_buf       *partial;
	rx_packet_handler_t  handler;
	void                *udata;
};

/*****************************************************************************/

#define RX_PACKET(rx) \
	((struct rx_packet *) ((rx)->udata))

/*****************************************************************************/

static void null_packet_handler (void *udata, GtPacket *packet)
{
	gt_packet_free (packet);
}

static BOOL rx_packet_init (struct rx_layer *rx, void *udata)
{
	struct rx_packet *rx_packet;

	if (!(rx_packet = NEW (struct rx_packet)))
		return FALSE;

	rx->udata = rx_packet;

	/* discard the packets received */
	rx_packet->handler = null_packet_handler;
	return TRUE;
}

static void rx_packet_destroy (struct rx_layer *rx)
{
	struct rx_packet *rx_packet = RX_PACKET(rx);

	io_buf_free (rx_packet->partial);
	FREE (rx_packet);
}

static void rx_packet_disable (struct rx_layer *rx)
{
	/* NOP: lower layer will stop sending us data -- it'd better
	 * stop now */
}

static void rx_packet_enable (struct rx_layer *rx)
{
	/* NOP: the lower layer will start sending us data again */
}

/*****************************************************************************/

static uint32_t get_len_to_read (struct io_buf *partial, size_t avail, 
                                 uint32_t *r_packet_size)
{
	uint32_t packet_size;
	uint32_t len;
	uint32_t payload_len;

	*r_packet_size = 0;

	/*
	 * Check if the partial packet hasn't completed reading
	 * the header yet. If it hasn't, we need to read up until
	 * we reach the payload length (that starts at byte 19)
	 * before we know how big the packet is.
	 */
	if (io_buf_len (partial) < GNUTELLA_HDR_LEN)
	{
		/* not enough data in the recv buffer for the complete
		 * Gnutella header */
		if (io_buf_len (partial) + avail < GNUTELLA_HDR_LEN)
			return avail;

		/* get the remaining bytes to complete the header */
		return GNUTELLA_HDR_LEN - io_buf_len (partial);
	}

	/* 
	 * The partial packet is now at least 23 bytes. Look in the header for the
	 * payload length so we know how much we need to read before the packet is
	 * complete.
	 */
	payload_len = get_payload_len (partial->data);
	packet_size = payload_len + GNUTELLA_HDR_LEN;

	/*
	 * Check for wraparound, and reset the packet size its payload len.
	 * Its likely we've experienced a protocol de-sync here, so even
	 * though the size is wrong its important that it be large so we
	 * can detect it later.
	 */
	if (packet_size < GNUTELLA_HDR_LEN)
		packet_size = payload_len;

	len = packet_size - io_buf_len (partial);
	assert (len >= 0); /* otherwise wouldve flushed last call */

	if (len > avail)
		len = avail;

	*r_packet_size = packet_size;
	return len;
}

static GtPacket *make_packet (struct rx_layer *rx, struct rx_packet *rx_packet,
                              size_t packet_size)
{
	GtPacket      *packet;
	struct io_buf *pbuf = rx_packet->partial;

	assert (io_buf_len (pbuf) == packet_size);

	/* construct a complete packet from the data in the binary buf */
	packet = gt_packet_unserialize (pbuf->data, packet_size);

	/* 
	 * TODO: gt_packet_unserialize() currently copies the data, but should
	 * really take ownership of the data in place. That would also allow
	 * the null terminator in the io_buf to terminate the last string in
	 * the packet, if any.
	 */
	io_buf_free (pbuf);
	rx_packet->partial = NULL;

	if (!packet)
	{
		gt_rx_stack_abort (rx->stack);
		return NULL;
	}

	return packet;
}

/*
 * We must read all the data the lower layer has sent and buffer
 * partial packets, because otherwise we would poll the CPU if the
 * packet were bigger than the buffer size of one of the layer
 * layers under this one.
 */
static BOOL read_packet (struct rx_layer *rx, struct io_buf *io_buf,
                         size_t avail, GtPacket **ret)
{
	struct rx_packet   *rx_packet   = rx->udata;
	struct io_buf      *partial;
	size_t              total;
	uint32_t            packet_size = 0;
	uint32_t            len_to_read;

	if (!rx_packet->partial && 
	    !(rx_packet->partial = io_buf_new (GNUTELLA_HDR_LEN)))
	{
		gt_rx_stack_abort (rx->stack);
		return FALSE;
	}

	partial = rx_packet->partial;

	len_to_read = get_len_to_read (partial, avail, &packet_size);

	/* ensure the packet has enough room */
	if (!io_buf_resize (partial, io_buf_len (partial) + len_to_read))
	{
		gt_rx_stack_abort (rx->stack);
		return FALSE;
	}

	/* TODO: make this a per-rxstack variable, because we could want a larger
	 * packet size for doing browsing, for example */
	if (packet_size >= GT_PACKET_MAX)
	{
		GT->dbg (GT, "received too large packet(%d)", packet_size);
		gt_rx_stack_abort (rx->stack);
		return FALSE;
	}

	total = io_buf_copy (partial, io_buf, len_to_read);

	assert (total <= avail);
	assert (io_buf_len (partial) == partial->w_offs);

	avail -= total;

	/* we didn't read a complete packet, so stop reading: the
	 * packet-reading loop was unable to satisfy us */
	if (io_buf_len (partial) < GNUTELLA_HDR_LEN)
	{
		assert (avail == 0);
		assert (packet_size == 0);
		return FALSE;
	}

	/* 
	 * Restart processing at higher level to read packet data:
	 * this is done to simplify the processing a bit and so we
	 * don't risk reading part of some other packet's data. 
	 */
	if (packet_size == 0)
		return TRUE;

	assert (packet_size >= GNUTELLA_HDR_LEN);
	assert (io_buf_len (partial) >= GNUTELLA_HDR_LEN);

	/* 
	 * If we read enough for the whold payload length, we finished reading a
	 * complete packet. Otherwise, we've read everything that was available
	 * but didn't find a packet.
	 */
	if (io_buf_len (partial) >= packet_size)
	{
		*ret = make_packet (rx, rx_packet, packet_size);
		return (*ret ? TRUE : FALSE);
	}

	/* 
	 * Did partial packet read: there must be no more data, otherwise we
	 * would've put it inside our current incomplete packet.
	 */
	assert (avail == 0);

	return FALSE;
}

/*
 * Receive a message buffer from the lower layer, parse it into as many
 * packets as it contains, and pass those to our handler function.
 * 
 * This is meant to be the top layer of the rx stack only.
 *
 * TODO: A handler could be implemented as another layer, should think
 * about this approach instead.
 */
static void rx_packet_recv (struct rx_layer *rx, struct io_buf *io_buf)
{
	GtPacket          *packet = NULL;
	size_t             avail;
	struct rx_packet  *rx_packet;

	rx_packet = RX_PACKET(rx);

	while (rx->enabled && (avail = io_buf_read_avail (io_buf)) > 0)
	{
		if (!read_packet (rx, io_buf, avail, &packet))
		{
			assert (packet == NULL);
			break;
		}

		if (packet)
		{
			(*rx_packet->handler) (rx_packet->udata, packet);

			/* freeing the packet here means the callback must make
			 * its own provisions for storing the packet's data */
			gt_packet_free (packet);

			packet = NULL;
		}
	}

	io_buf_free (io_buf);
}
                            
/*****************************************************************************/

struct rx_layer_ops rx_packet_ops =
{
	rx_packet_init,
	rx_packet_destroy,
	rx_packet_enable,
	rx_packet_disable,
	rx_packet_recv,
};

void rx_packet_set_handler (struct rx_layer *rx, 
                            rx_packet_handler_t handler,
                            void *udata)
{
	struct rx_packet *rx_packet = RX_PACKET(rx);

	rx_packet->handler = handler;
	rx_packet->udata   = udata;
}
